#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

__author__ = "Olof Svensson"
__contact__ = "svensson@esrf.eu"
__copyright__ = "ESRF"
__updated__ = "2020-12-17"

import os
import numpy
import pprint
import textwrap
import matplotlib
import matplotlib.cm
import matplotlib.pyplot as plt

from edna2.tasks.AbstractTask import AbstractTask

from edna2.utils import UtilsLogging
from edna2.utils import UtilsDetector


logger = UtilsLogging.getLogger()

class DozorM(AbstractTask):  # pylint: disable=too-many-instance-attributes
    """
    The DozorM is responsible for executing the 'dozorm' program.
    """

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                # "listDozorAllFile": {
                #     "type": "array",
                #     "items": {"type": "string"},
                # },
                "name_template_scan": {"type": "string"},
                "detectorType": {"type": "string"},
                "beamline": {"type": "string"},
                "detector_distance": {"type": "number"},
                "wavelength": {"type": "number"},
                "orgx": {"type": "number"},
                "orgy": {"type": "number"},
                "number_row": {"type": "number"},
                "number_images": {"type": "number"},
                "isZigZag": {"type": "boolean"},
                "step_h": {"type": "number"},
                "step_v": {"type": "number"},
                "beam_shape": {"type": "string"},
                "beam_h": {"type": "number"},
                "beam_v": {"type": "number"},
                "number_apertures": {"type": "integer"},
                "aperture_size": {"type": "string"},
                "reject_level": {"type": "integer"},
                "isHorizontalScan": {"type": "boolean"},
                "number_scans": {"type": "integer"}
            }
        }

    def getOutDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "dozorMap": {"type": "string"},
                "listPositions":  {
                    "type": "array",
                    "items": {"type": "object"}
                },
            },
        }

    def run(self, inData):
        outData = {}
        commands = self.generateCommands(inData)
        with open(str(self.getWorkingDirectory() / 'dozorm.dat'), 'w') as f:
            f.write(commands)
        commandLine = "dozorm dozorm.dat"
        logPath = self.getWorkingDirectory() / 'dozorm.log'
        self.runCommandLine(commandLine, logPath=logPath)
        outData = self.parseOutput(self.getWorkingDirectory(), logPath)
        return outData

    def generateCommands(self, inData):
        """
        This method creates the input file for dozorm
        """
        detectorType = inData['detectorType']
        nx = UtilsDetector.getNx(detectorType)
        ny = UtilsDetector.getNy(detectorType)
        pixelSize = UtilsDetector.getPixelsize(detectorType)
        nameTemplateScan = self.getWorkingDirectory() / "dozorm_00?"
        os.symlink(inData["dozorAllFile"], str(self.getWorkingDirectory() / "dozorm_001"))
        firstScanNumber = 1
        if inData.get('isHorizontalScan', True):
            meshDirect = "-h"
        else:
            meshDirect = "-v"
        command = '!\n'
        command += 'detector {0}\n'.format(detectorType)
        command += 'nx %d\n' % nx
        command += 'ny %d\n' % ny
        command += 'pixel %f\n' % pixelSize
        command += 'detector_distance {0}\n'.format(inData['detector_distance'])
        command += 'X-ray_wavelength {0}\n'.format(inData['wavelength'])
        command += 'orgx {0}\n'.format(inData['orgx'])
        command += 'orgy {0}\n'.format(inData['orgy'])
        command += 'number_row {0}\n'.format(inData['number_row'])
        command += 'number_images {0}\n'.format(inData['number_images'])
        command += 'mesh_direct {0}\n'.format(meshDirect)
        command += 'step_h {0}\n'.format(inData['step_h'])
        command += 'step_v {0}\n'.format(inData['step_v'])
        command += 'beam_shape {0}\n'.format(inData['beam_shape'])
        command += 'beam_h {0}\n'.format(inData['beam_h'])
        command += 'beam_v {0}\n'.format(inData['beam_v'])
        command += 'number_apertures {0}\n'.format(inData['number_apertures'])
        command += 'aperture_size {0}\n'.format(inData['aperture_size'])
        command += 'reject_level {0}\n'.format(inData['reject_level'])
        command += 'name_template_scan {0}\n'.format(nameTemplateScan)
        command += 'number_scans {0}\n'.format(inData['number_scans'])
        command += 'first_scan_number {0}\n'.format(firstScanNumber)
        command += 'end\n'
        # logger.debug('command: {0}'.format(command))
        return command

    @staticmethod
    def parseOutput(workingDir, logPath):
        outData = {}
        pathDozormMap = workingDir / "dozorm_001.map"
        if pathDozormMap.exists():
            dictMap = DozorM.parseMap(pathDozormMap)
            nx = dictMap["nx"]
            ny = dictMap["ny"]
            listPositions = DozorM.parseDozormLogFile(logPath)
            # Fix for problem with 1D scans
            listPositions = DozorM.check1Dpositions(listPositions, nx, ny)
            crystalMapPath = DozorM.makeCrystalPlot(dictMap["crystal"], workingDir)
            if nx != 1 and ny != 1:
                imageNumberMapPath = DozorM.makeImageNumberMap(dictMap["imageNumber"], workingDir)
            else:
                imageNumberMapPath = None
            outData = {
                "dozorMap": str(pathDozormMap),
                "listPositions": listPositions,
                "nx": nx,
                "ny": ny,
                "score": dictMap["score"],
                "crystal": dictMap["crystal"],
                "imageNumber": dictMap["imageNumber"],
                "crystalMapPath": crystalMapPath,
                "imageNumberMapPath": imageNumberMapPath
            }
        return outData

    @staticmethod
    def parseDozormLogFile(logPath):
        listPositions = []
        #    Total N.of crystals in Loop =  2
        # Cryst Aperture Central  Coordinate    Int/Sig N.of Images Score  Helic   Start     Finish   Int/Sig
        # number size     image      X    Y            All dX dY    sum           x     y     x    y   helical
        # --------------------------------------------------------------------------------------------------
        #     1   100.0     260    8.1   15.1  545.7  12   3   5  1242.5   NO
        #     2   100.0      82    9.9    5.0  354.0   7   3   3   670.6   NO
        #     3    10.0     183    2.7   10.8   31.2   3   2   2   169.1   NO
        #     4    10.0     170   10.8    9.7   26.7   3   2   3   128.6  YES    11    10    12    11   100.7

        with open(str(logPath)) as fd:
            listLogLines = fd.readlines()
        doParseLine = False
        for line in listLogLines:
            if line.startswith("------"):
                doParseLine = True
            elif doParseLine:
                listValues = line.split()
                try:
                    iOverSigma = float(listValues[5])
                except:
                    iOverSigma = listValues[5]
                position = {
                    "number": int(listValues[0]),
                    "apertureSize": str(int(float(listValues[1]))),
                    "imageNumber": int(listValues[2]),
                    "xPosition": float(listValues[3]),
                    "yPosition": float(listValues[4]),
                    "iOverSigma": iOverSigma,
                    "numberOfImagesTotal": int(listValues[6]),
                    "numberOfImagesTotalX": int(listValues[7]),
                    "numberOfImagesTotalY": int(listValues[8]),
                    "score": float(listValues[9]),
                    "helical": listValues[10] == 'YES'
                }
                if position["helical"]:
                    position["helicalStartX"] = listValues[11]
                    position["helicalStartY"] = listValues[12]
                    position["helicalStopX"] = listValues[13]
                    position["helicalStopY"] = listValues[14]
                    position["helicalIoverSigma"] = listValues[15]
                listPositions.append(position)
        return listPositions

    @staticmethod
    def makeCrystalPlot(arrayCrystal, workingDirectory, debug=False):
        npArrayCrystal = numpy.array(arrayCrystal)
        ySize, xSize = npArrayCrystal.shape
        if xSize == 1:
            # Vertical line scan - transpose the matrix
            npArrayCrystal = numpy.transpose(npArrayCrystal)
            ySize, xSize = npArrayCrystal.shape
        npArrayCrystalAbs = numpy.abs(npArrayCrystal)
        # minValue = numpy.amin(npArrayCrystal)
        # newZeroValue = minValue - 1
        # npArrayCrystal = numpy.where(npArrayCrystal == 0.0, newZeroValue, npArrayCrystal)

        maxSize = max(xSize, ySize)
        if maxSize < 10:
            fontSize = 12
            dpi = 75
        elif maxSize < 50:
            fontSize = 8
            dpi = 100
        else:
            fontSize =5
            dpi = 150

        font = {'family': 'normal',
                'weight': 'normal',
                'size': fontSize}

        matplotlib.rc('font', **font)

        fig, ax = plt.subplots()

        im = ax.imshow(
            npArrayCrystalAbs,
            cmap=matplotlib.cm.Spectral
        )

        ax.set_xticks(numpy.arange(len(range(xSize))))
        ax.set_yticks(numpy.arange(len(range(ySize))))

        ax.set_xticklabels(list(range(1, xSize+1)))
        ax.set_yticklabels(list(range(1, ySize+1)))

        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
                 rotation_mode="anchor")

        # Loop over data dimensions and create text annotations.
        for i in range(ySize):
            for j in range(xSize):
                if abs(npArrayCrystal[i, j]) > 0.001:
                    text = ax.text(j, i, npArrayCrystal[i, j],
                                   ha="center", va="center", color="b")

        ax.set_title("Crystal map")
        fig.tight_layout(pad=2)
        w, h = fig.get_size_inches()
        x1, x2 = ax.get_xlim()
        y1, y2 = ax.get_ylim()
        x1 = float(x1)
        x2 = float(x2)
        y1 = float(y1)
        y2 = float(y2)
        fig.set_size_inches(w + 2, abs(y2 - y1) / (x2 - x1) * w + 2)
        if debug:
            plt.show()
        crystalMapPath = os.path.join(workingDirectory, "crystalMap.png")
        plt.savefig(crystalMapPath, dpi=dpi)

        return crystalMapPath

    @staticmethod
    def makeImageNumberMap(arrayImageNumber, workingDirectory, debug=False):
        npImageNumber = numpy.array(arrayImageNumber)
        npArrayImageNumber = numpy.zeros(npImageNumber.shape)
        ySize, xSize = npImageNumber.shape

        maxSize = max(xSize, ySize)
        if maxSize < 10:
            fontSize = 12
            dpi = 75
        elif maxSize < 50:
            fontSize = 8
            dpi = 100
        else:
            fontSize =5
            dpi = 150

        font = {'family': 'normal',
                'weight': 'normal',
                'size': fontSize}

        matplotlib.rc('font', **font)

        fig, ax = plt.subplots()
        im = ax.imshow(
            npArrayImageNumber,
            cmap=matplotlib.cm.Greys
        )

        ax.set_xticks(numpy.arange(len(range(xSize))))
        ax.set_yticks(numpy.arange(len(range(ySize))))

        ax.set_xticklabels(list(range(1, xSize+1)))
        ax.set_yticklabels(list(range(1, ySize+1)))

        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
                 rotation_mode="anchor")

        # Loop over data dimensions and create text annotations.
        for i in range(ySize):
            for j in range(xSize):
                text = ax.text(j, i, arrayImageNumber[i][j],
                               ha="center", va="center", color="b")

        ax.set_title("Image numbers")
        fig.tight_layout(pad=2)
        w, h = fig.get_size_inches()
        x1, x2 = ax.get_xlim()
        y1, y2 = ax.get_ylim()
        x1 = float(x1)
        x2 = float(x2)
        y1 = float(y1)
        y2 = float(y2)
        fig.set_size_inches(w + 2, abs(y2 - y1) / (x2 - x1) * w + 2)
        if debug:
            plt.show()
        imageNumberPath = os.path.join(workingDirectory, "imageNumber.png")
        plt.savefig(imageNumberPath, dpi=dpi)

        return imageNumberPath

    @staticmethod
    def parseMatrix(index, listLines, spacing, isFloat=True):
        arrayValues = []
        # Parse matrix - starts and ends with "---" line
        while not listLines[index].startswith("-------"):
            index += 1
        index += 1
        while not listLines[index].startswith("-------"):
            listScores = textwrap.wrap(listLines[index][5:], spacing)
            if isFloat:
                listScores = list(map(float, listScores))
            else:
                listScores = list(map(int, listScores))
            arrayValues.append(listScores)
            index += 1
        index += 1
        return index, arrayValues



    @staticmethod
    def parseMap(mapPath):
        with open(str(mapPath)) as fd:
            listLines = fd.readlines()
        # Parse map dimensions
        index = 1
        nx, ny = map(int, listLines[index].split())
        # Parse scores
        index, arrayScore = DozorM.parseMatrix(index, listLines, spacing=6, isFloat=True)
        # Parse crystals
        index, arrayCrystal = DozorM.parseMatrix(index, listLines, spacing=4, isFloat=False)
        # Parse image number
        index, arrayImageNumber = DozorM.parseMatrix(index, listLines, spacing=5, isFloat=False)
        dictMap = {
            "nx": nx,
            "ny": ny,
            "score": arrayScore,
            "crystal": arrayCrystal,
            "imageNumber": arrayImageNumber
        }
        return dictMap

    @staticmethod
    def updateMeshPositions(meshPositions, arrayScore):
        newMeshPositions = []
        for position in meshPositions:
            # pprint.pprint(position)
            indexY = position["indexY"]
            indexZ = position["indexZ"]
            # print(indexY, indexZ)
            dozormScore = arrayScore[indexZ][indexY]
            dozorScore = position["dozor_score"]
            # print(dozorScore, dozormScore)
            newPosition = dict(position)
            newPosition["dozor_score_orig"] = dozorScore
            newPosition["dozor_score"] = dozormScore
            newMeshPositions.append(newPosition)
        return newMeshPositions

    @staticmethod
    def check1Dpositions(listPositions, nx, ny):
        newListPositions = []
        for position in listPositions:
            newPosition = dict(position)
            if nx == 1:
                newPosition["xPosition"] = 1.0
            if ny == 1:
                newPosition["yPosition"] = 1.0
            newListPositions.append(newPosition)
        return newListPositions



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "10/05/2019"

import json
import shutil

from edna2.tasks.AbstractTask import AbstractTask
from edna2.tasks.ISPyBTasks import GetListAutoprocessingResults


class FindHklAsciiForMerge(AbstractTask):
    """
    This task receives a list of data collection IDs and returns a
    json schema for EXI2
    """

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "token": {"type": "string"},
                "proposal": {"type": "string"},
                "dataCollectionId": {
                    "type": "array",
                    "items": {
                        "type": "integer",
                    }
                }
            }
        }

    # def getOutDataSchema(self):
    #     return {
    #         "type": "object",
    #         "required": ["dataForMerge"],
    #         "properties": {
    #             "dataForMerge": {
    #                 "type": "object",
    #                 "items": {
    #                     "type": "object",
    #                     "properties": {
    #                         "spaceGroup": {"type": "string"}
    #                     }
    #                 }
    #             }
    #         }
    #     }

    def run(self, inData):
        urlError = None
        token = inData['token']
        proposal = inData['proposal']
        listDataCollectionId = inData['dataCollectionId']
        inDataGetListAutoprocessingResults = {
            'token': token,
            'proposal': proposal,
            'dataCollectionId': listDataCollectionId
        }
        getListAutoprocessingResults = GetListAutoprocessingResults(
            inData=inDataGetListAutoprocessingResults
        )
        getListAutoprocessingResults.execute()
        outDataAutoprocessing = getListAutoprocessingResults.outData
        if 'error' in outDataAutoprocessing:
            urlError = outDataAutoprocessing['error']
        else:
            index = 1
            properties = {}
            listOrder = []
            for dataCollection in outDataAutoprocessing['dataCollection']:
                dataCollectionId = dataCollection['dataCollectionId']
                dictEntry = {}
                listEnumNames = []
                listEnumValues = []
                proteinAcronym = None
                blSampleName = None
                if 'error' in dataCollection['autoprocIntegration']:
                    urlError = dataCollection['autoprocIntegration']['error']
                else:
                    for autoProcResult in dataCollection['autoprocIntegration']:
                        if proteinAcronym is None:
                            proteinAcronym = autoProcResult['Protein_acronym']
                            blSampleName = autoProcResult['BLSample_name']
                        for autoProcAttachment in autoProcResult['autoprocAttachment']:
                            if 'XDS_ASCII' in autoProcAttachment['fileName']:
                                fileName = autoProcAttachment['fileName']
                                program = autoProcResult['v_datacollection_processingPrograms']
                                attachmentId = autoProcAttachment['autoProcProgramAttachmentId']
                                enumName = '{0:30s} {1}'.format(program, fileName)
                                listEnumNames.append(enumName)
                                enumValue = attachmentId
                                listEnumValues.append(enumValue)
                if urlError is None:
                    entryKey = 'hkl_' + str(dataCollectionId)
                    if entryKey not in properties:
                        dictEntry['title'] = 'Select HKL for data Collection #{0} {2} {1}-{2}'.format(
                            index,
                            proteinAcronym,
                            blSampleName
                        )
                        dictEntry['enum'] = listEnumValues
                        dictEntry['enumNames'] = listEnumNames
                        properties[entryKey] = dictEntry
                        listOrder.append(entryKey)
                    entryKey = 'minimum_I/SIGMA_' + str(dataCollectionId)
                    if entryKey not in properties:
                        # Minimum sigma
                        dictEntry = {
                            'integer': 'string',
                            'type': 'string',
                            'title': 'minimum_I/SIGMA for data Collection #{0} {2} {1}-{2}'.format(
                                index,
                                proteinAcronym,
                                blSampleName
                            )
                        }
                        properties[entryKey] = dictEntry
                        listOrder.append(entryKey)
                        index += 1
        if urlError is None:
            schema = {
                'properties': properties,
                'type': 'object',
                'title': 'User input needed'
            }
            uiSchema = {
                'ui:order': listOrder
            }
            outData = {
                "schema": schema,
                "uiSchema": uiSchema
            }
        else:
            outData = {
                'error': urlError
            }
        return outData


class FindPipelineForMerge(AbstractTask):
    """
    This task receives a list of data collection IDs and returns a
    json schema for EXI2
    """

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "token": {"type": "string"},
                "proposal": {"type": "string"},
                "dataCollectionId": {
                    "type": "array",
                    "items": {
                        "type": "integer",
                    }
                }
            }
        }

    # def getOutDataSchema(self):
    #     return {
    #         "type": "object",
    #         "required": ["dataForMerge"],
    #         "properties": {
    #             "dataForMerge": {
    #                 "type": "object",
    #                 "items": {
    #                     "type": "object",
    #                     "properties": {
    #                         "spaceGroup": {"type": "string"}
    #                     }
    #                 }
    #             }
    #         }
    #     }

    def run(self, inData):
        urlError = None
        token = inData['token']
        proposal = inData['proposal']
        listDataCollectionId = inData['dataCollectionId']
        inDataGetListAutoprocessingResults = {
            'token': token,
            'proposal': proposal,
            'dataCollectionId': listDataCollectionId
        }
        getListAutoprocessingResults = GetListAutoprocessingResults(
            inData=inDataGetListAutoprocessingResults
        )
        getListAutoprocessingResults.execute()
        outDataAutoprocessing = getListAutoprocessingResults.outData
        if 'error' in outDataAutoprocessing:
            urlError = outDataAutoprocessing['error']
        else:
            index = 1
            properties = {}
            listOrder = []
            dictEntry = {}
            for dataCollection in outDataAutoprocessing['dataCollection']:
                dataCollectionId = dataCollection['dataCollectionId']
                listEnumValues = []
                proteinAcronym = None
                blSampleName = None
                if 'error' in dataCollection['autoprocIntegration']:
                    urlError = dataCollection['autoprocIntegration']['error']
                else:
                    for autoProcResult in dataCollection['autoprocIntegration']:
                        if len(autoProcResult['autoprocAttachment']) > 0:
                            if proteinAcronym is None:
                                proteinAcronym = autoProcResult['Protein_acronym']
                                blSampleName = autoProcResult['BLSample_name']
                            if '1' in autoProcResult['anomalous']:
                                anom = True
                            else:
                                anom = False
                            for autoProcAttachment in autoProcResult['autoprocAttachment']:
                                if 'XDS_ASCII' in autoProcAttachment['fileName']:
                                    fileName = autoProcAttachment['fileName']
                                    program = autoProcResult['v_datacollection_processingPrograms']
                                    attachmentId = autoProcAttachment['autoProcProgramAttachmentId']
                                    if anom:
                                        entryKey = program + '_anom'
                                    else:
                                        entryKey = program + '_noanom'
                                    if entryKey not in dictEntry:
                                        dictEntry[entryKey] = []
                                    dictEntry[entryKey].append({'id': attachmentId, 'fileName': fileName})
        if urlError is None:
            listEnumNames = []
            dictInput = {}
            for entryKey, listAttachment in dictEntry.items():
                if len(listAttachment) == len(outDataAutoprocessing['dataCollection']):
                    listEnumNames.append(entryKey)
                    dictInput[entryKey] = listAttachment
                    index += 1
            if len(listEnumNames) > 0:
                dictSchema = {
                    'title': 'Select processing pipeline for data Collection {0}-{1}'.format(
                        proteinAcronym,
                        blSampleName
                    ),
                    'type': 'string',
                    'enum': listEnumNames,
                    'enumNames': listEnumNames
                }
                key = "pipeline"
                properties[key] = dictSchema
                listOrder.append(key)
                # Minimum sigma
                dictSchema = {
                    'integer': 'string',
                    'type': 'string',
                    'title': 'minimum_I/SIGMA for data Collection {0}-{1}'.format(
                        proteinAcronym,
                        blSampleName
                    )
                }
                key = 'minimum_I/SIGMA'
                properties[key] = dictSchema
                listOrder.append(key)
        if urlError is None:
            schema = {
                'properties': properties,
                'type': 'object',
                'title': 'User input needed'
            }
            uiSchema = {
                'ui:order': listOrder
            }
            outData = {
                'schema': {
                    "schema": schema,
                    "uiSchema": uiSchema
                },
                'input': dictInput
            }
        else:
            outData = {
                'error': urlError
            }
        return outData


class MergeUtls(AbstractTask):
    """
    This task will run the Merge_utls.py program written by Shibom Basu
    """

    def run(self, inData):
        listHklLp = inData['listHklLp']
        workingDir = self.getWorkingDirectory()
        index = 1
        for hklLp in listHklLp:
            dataDir = workingDir / "data{0}".format(index)
            dataDir.mkdir(exist_ok=True)
            shutil.copy(hklLp['hkl'], str(dataDir / 'XDS_ASCII.HKL'))
            index += 1
        commandLine = 'Merge_utls.py --root {0} --expt serial-xtal'.format(str(workingDir))
        self.runCommandLine(commandLine, logPath=None)
        # Find Mergeing_results.json
        resultPath = self.getWorkingDirectory() / 'adm_serial-xtal' / 'adm_3' / 'Mergeing_results.json'
        if resultPath.exists():
            with open(str(resultPath)) as f:
                mergeResult = json.loads(f.read())
        outData = {'mergeResult': mergeResult}
        return outData



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "14/04/2020"


import numpy as np

from edna2.tasks.AbstractTask import AbstractTask
from edna2.tasks.SubWedgeAssembly import SubWedgeAssembly
from edna2.tasks.XDSTasks import XDSIndexAndIntegration

from edna2.utils import UtilsImage

class RadiationDamageProcessing(AbstractTask):
    """
    This task receives a list of images or data collection ids and
    returns result of indexing
    """

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "dataCollectionId": {"type": "integer"},
                "imagePath": {
                    "type": "array",
                    "items": {
                        "type": "string",
                    },
                },
            },
        }

    def run(self, inData):
        outData = {}
        # First get the list of subWedges
        if "subWedge" in inData:
            list_sub_wedge = inData["subWedge"]
            directory_prefix = ""
        else:
            first_image = inData["imagePath"][0]
            last_image = inData["imagePath"][-1]
            prefix = UtilsImage.getPrefix(first_image)
            first_image_number = UtilsImage.getImageNumber((first_image))
            last_image_number = UtilsImage.getImageNumber((last_image))
            directory_prefix = "{0}_{1}_{2}".format(
                prefix, first_image_number, last_image_number
            )
            list_sub_wedge = self.getListSubWedge(inData)
        indata_xds_integration = {
            "subWedge": list_sub_wedge
        }
        xds_integration = XDSIndexAndIntegration(
            inData=indata_xds_integration,
            workingDirectorySuffix=directory_prefix,
        )
        xds_integration.execute()
        outData = xds_integration.outData
        return outData




    def getListSubWedge(self, inData):
        list_sub_wedges = None
        sub_wedge_assembly = SubWedgeAssembly(inData=inData)
        sub_wedge_assembly.execute()
        if sub_wedge_assembly.isSuccess():
            list_sub_wedges = sub_wedge_assembly.outData["subWedge"]
        return list_sub_wedges



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "29/03/2022"

from edna2.tasks.AbstractTask import AbstractTask
from edna2.tasks.ReadImageHeader import ReadImageHeader

from edna2.utils import UtilsImage
from edna2.utils import UtilsLogging
from edna2.utils import UtilsSubWedge

logger = UtilsLogging.getLogger()


class SubWedgeAssembly(AbstractTask):

    def run(self, in_data):
        sub_wedge_merge = []
        fast_characterisation = {}
        force_zero_rotation_axis_start = False
        if "fastCharacterisation" in in_data:
            fast_characterisation = in_data["fastCharacterisation"]
            force_zero_rotation_axis_start = fast_characterisation.get("forceZeroRotationAxisStart", False)
            is_fast_characterisation = True
            list_subwedge_angles = fast_characterisation["listSubWedgeAngles"]
            no_images_in_subwedge = fast_characterisation["noImagesInSubWedge"]
            first_image_path = fast_characterisation["firstImagePath"]
            image_number = 1
            template = first_image_path.replace("0001", "{0:04d}")
            list_image_path = []
            for subwedge_angle in list_subwedge_angles:
                for sub_wedge_image_number in range(no_images_in_subwedge):
                    image_path = template.format(image_number)
                    list_image_path.append(image_path)
                    image_number += 1
        elif "imagePath" in in_data:
            list_image_path = in_data["imagePath"]
            is_fast_characterisation = False
            list_subwedge_angles = None
            no_images_in_subwedge = None
        else:
            raise RuntimeError("Neither 'imagePath' nor 'fastCharacterisation' in input data.")
        sub_wedge_number = 1
        input_read_image_header = {
            "imagePath": list_image_path
        }
        read_image_header = ReadImageHeader(inData=input_read_image_header)
        read_image_header.execute()
        if read_image_header.isSuccess():
            list_subwedge = read_image_header.outData["subWedge"]
            global_axis_start = None
            if force_zero_rotation_axis_start:
                for subwedge in list_subwedge:
                    axis_start = subwedge["experimentalCondition"]["goniostat"]["rotationAxisStart"]
                    if global_axis_start is None or global_axis_start > axis_start:
                        global_axis_start = axis_start
            for index_subwedge, subwedge in enumerate(list_subwedge):
                if is_fast_characterisation:
                    # Correct for MASSIF 1 angle bug
                    subwedge["experimentalCondition"]["goniostat"]["rotationAxisStart"] = index_subwedge * 0.1
                    subwedge["experimentalCondition"]["goniostat"]["rotationAxisEnd"] = index_subwedge * 0.1 + 0.1
                    subwedge["experimentalCondition"]["goniostat"]["oscillationWidth"] = 0.1
                    # Modify the start angle
                    index_angle = int(index_subwedge / no_images_in_subwedge)
                    angle_subwedge = list_subwedge_angles[index_angle] - index_angle
                    goniostat = subwedge["experimentalCondition"]["goniostat"]
                    if force_zero_rotation_axis_start:
                        goniostat["rotationAxisStart"] -= global_axis_start
                    goniostat["rotationAxisStart"] = (goniostat["rotationAxisStart"] + angle_subwedge) % 360
                    goniostat["rotationAxisEnd"] = (goniostat["rotationAxisEnd"] + angle_subwedge) % 360
                else:
                    subwedge["subWedgeNumber"] = index_subwedge + 1
            sub_wedge_merge = UtilsSubWedge.subWedgeMerge(list_subwedge)
        out_data = {
            "subWedge": sub_wedge_merge
        }
        return out_data



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/03/2022"

import json

from edna2.utils import UtilsLogging

logger = UtilsLogging.getLogger()


def subWedgeMerge(inData):
    list_result_sub_wedge_merge = []
    # 1. Sort the incoming subwedges in groupes of similar experimental conditions
    list_sub_wedge_sorted = sortSubWedgesOnExperimentalCondition(inData)

    # 2. Merge subwedges which are adjascent with respect to the orientation matrix
    for list_sub_wedge in list_sub_wedge_sorted:
        list_sub_wedge_merged = mergeListOfSubWedgesWithAdjascentRotationAxis(
            list_sub_wedge
        )
        for index, sub_wedge in enumerate(list_sub_wedge_merged):
            sub_wedge["subWedgeNumber"] = index + 1
        list_result_sub_wedge_merge += list_sub_wedge_merged
    return list_result_sub_wedge_merge




def compareTwoValues(value1, value2, tolerance=0.001):
    """
    This method compares two values which can be of type double, integer or string.
    The types of the two values two be compared must be identical.
    If the value is of type double, an optional tolerance can be passed as an
    argument to the method. If both values are None the methods returns True.
    """
    return_value = False
    if value1 is None and value2 is None:
        return_value = True
    elif type(value1) != type(value2):
        error_message = "Types of values different : value1=%r, value2=%r" % (
            type(value1),
            type(value2),
        )
        logger.error(error_message)
        raise RuntimeError(error_message)
    elif type(value1) == float:
        if abs(value1 - value2) < tolerance:
            return_value = True
    elif type(value1) == int:
        if value1 == value2:
            return_value = True
    elif type(value1) == str:
        if value1 == value2:
            return_value = True
    else:
        error_message = "Unknown value type : %r for value %r" % (type(value1), value1)
        logger.error(error_message)
        raise RuntimeError(error_message)
    return return_value


def compareTwoParameters(exp_cond_1, exp_cond_2, type1, type2, tolerance=0.001):
    return_value = True
    if type1 in exp_cond_1 and type1 in exp_cond_2:
        if type2 in exp_cond_1[type1] and type2 in exp_cond_2[type1]:
            value1 = exp_cond_1[type1][type2]
            value2 = exp_cond_2[type1][type2]
            return_value = compareTwoValues(value1, value2, tolerance=tolerance)
    return return_value


def isSameExperimentalCondition(exp_cond_1, exp_cond_2):
    """
    This method compares two experimental condition objects in order to verify if
    they can belong to the same sub wedge. The following parameters are checked:
    beam.exposureTime [s], tolerance 0.001
    beam.wavelength [A], tolerance 0.001
    detector.beamPositionX [mm], tolerance 0.1
    detector.beamPositionY [mm], tolerance 0.1
    detector.distance [mm], tolerance 0.1
    detector.name [string]
    detector.numberPixelX [int]
    detector.numberPixelY [int]
    detector.serialNumber [string]
    detector.twoTheta [degrees], tolerance 0.1
    goniostat.oscillationWidth [degrees], tolerance 0.001
    goniostat.rotationAxis [string]
    """
    return_value = True

    if not compareTwoParameters(exp_cond_1, exp_cond_2, "beam", "exposureTime"):
        return_value = False

    if not compareTwoParameters(exp_cond_1, exp_cond_2, "beam", "wavelength"):
        return_value = False

    if not compareTwoParameters(exp_cond_1, exp_cond_2, "detector", "distance"):
        return_value = False

    if not compareTwoParameters(exp_cond_1, exp_cond_2, "detector", "name"):
        return_value = False

    if not compareTwoParameters(
        exp_cond_1, exp_cond_2, "detector", "beamPositionX", tolerance=0.1
    ):
        return_value = False

    if not compareTwoParameters(
        exp_cond_1, exp_cond_2, "detector", "beamPositionY", tolerance=0.1
    ):
        return_value = False

    if not compareTwoParameters(exp_cond_1, exp_cond_2, "detector", "numberPixelX"):
        return_value = False

    if not compareTwoParameters(exp_cond_1, exp_cond_2, "detector", "numberPixelY"):
        return_value = False

    if not compareTwoParameters(exp_cond_1, exp_cond_2, "detector", "serialNumber"):
        return_value = False

    if not compareTwoParameters(exp_cond_1, exp_cond_2, "detector", "twoTheta"):
        return_value = False

    if not compareTwoParameters(exp_cond_1, exp_cond_2, "goniostat", "oscillationWidth"):
        return_value = False

    if not compareTwoParameters(exp_cond_1, exp_cond_2, "goniostat", "rotationAxis"):
        return_value = False

    return return_value


def isSameExperimentalConditionInSubWedge(subWedge1, subWedge2):
    """
    This method compares two XSDataSubWedge objects in order to verify if
    they can belong to the same sub wedge. The two experimental condition
    objects are compared using the method "isSameExperimentalCondition"
    """
    exp_cond_1 = subWedge1["experimentalCondition"]
    exp_cond_2 = subWedge2["experimentalCondition"]
    return isSameExperimentalCondition(exp_cond_1, exp_cond_2)


def sortIdenticalObjects(list_objects, method_comparison):
    """
    This method takes as input a list of objects with the same type.
    It returns a list containing list of identical objects. If an object
    is not identical to another it is returned in a list alone.
    """
    list_results = []
    number_of_objects = len(list_objects)
    if number_of_objects == 0:
        pass
    elif number_of_objects == 1:
        list_results.append(list_objects)
    else:
        # More than 1 object
        list_remaining_objects = list_objects
        while len(list_remaining_objects) > 0:
            index = 0
            current_object = list_remaining_objects[0]
            list_remaining_objects = list_remaining_objects[1:]
            list_sorted_objects = []
            list_sorted_objects.append(current_object)
            # print "Before loop: ", current_object, list_remaining_objects, iNumberOfRemainingObjects
            while index < len(list_remaining_objects):
                compare_object = list_remaining_objects[index]
                are_equal = method_comparison(current_object, compare_object)
                # print "   In the loop:", index, compare_object, are_equal, list_sorted_objects, list_remaining_objects
                if are_equal:
                    list_sorted_objects.append(compare_object)
                    list_remaining_objects_tmp = list_remaining_objects[0:index]
                    if index < len(list_remaining_objects):
                        list_remaining_objects_tmp.extend(
                            list_remaining_objects[index + 1 :]
                        )
                    list_remaining_objects = list_remaining_objects_tmp
                else:
                    index += 1
            list_results.append(list_sorted_objects)
    return list_results


def sortSubWedgesOnExperimentalCondition(listSubWedge):
    """
    This method sorts a list of sub wedges into a new list containing lists of
    sub wegdes with identical experimental conditions.
    """
    # Sort it
    listSubWedgeSorted = sortIdenticalObjects(
        listSubWedge, isSameExperimentalConditionInSubWedge
    )
    return listSubWedgeSorted


def mergeTwoSubWedgesAdjascentInRotationAxis(sub_wedge_1, sub_wedge_2):
    """
    This method takes as input two sub wedges and merges them to an unique subwedge, if possible,
    and returns the resulting merged sub wedge. If the merge is not possible a None is returned.
    """
    sub_wedge_merged = None
    # First check that the two sub wedges have identical experimental conditions
    if isSameExperimentalConditionInSubWedge(sub_wedge_1, sub_wedge_2):
        # Check if sub wedges are adjascent:
        rotation_axis_end_1 = sub_wedge_1["experimentalCondition"]["goniostat"][
            "rotationAxisEnd"
        ]
        rotation_axis_start_2 = sub_wedge_2["experimentalCondition"]["goniostat"][
            "rotationAxisStart"
        ]
        if compareTwoValues(rotation_axis_end_1, rotation_axis_start_2, 0.001):
            # Same sub wedge! Let's merge them
            sub_wedge_merged = json.loads(json.dumps(sub_wedge_1))
            sub_wedge_2 = json.loads(json.dumps(sub_wedge_2))
            rotation_axis_end_2 = sub_wedge_2["experimentalCondition"]["goniostat"][
                "rotationAxisEnd"
            ]
            sub_wedge_merged["experimentalCondition"]["goniostat"][
                "rotationAxisEnd"
            ] = rotation_axis_end_2
            for image in sub_wedge_2["image"]:
                sub_wedge_merged["image"].append(image)
    return sub_wedge_merged


def mergeListOfSubWedgesWithAdjascentRotationAxis(list_of_sub_wegdes):
    """
    This method merges sub wedges in a list if they are adjascent in phi.
    """
    # Copy the incoming list to a new list
    list_of_sub_wegdes_copy = json.loads(json.dumps(list_of_sub_wegdes))
    list_of_merged_sub_wedges = []
    if len(list_of_sub_wegdes_copy) == 0:
        pass
    elif len(list_of_sub_wegdes_copy) == 1:
        list_of_merged_sub_wedges = list_of_sub_wegdes_copy
    else:
        # First sort the list as function of rotation axis start
        list_of_sub_wegdes_copy.sort(
            key=lambda x: x["experimentalCondition"]["goniostat"]["rotationAxisStart"]
        )
        # Then loop through the subwedges and merge them if possible
        list_of_remaining_sub_wedges = list_of_sub_wegdes_copy
        current_sub_wedge = list_of_remaining_sub_wedges[0]
        list_of_remaining_sub_wedges = list_of_remaining_sub_wedges[1:]
        while len(list_of_remaining_sub_wedges) > 0:
            next_sub_wedge = list_of_remaining_sub_wedges[0]
            list_of_remaining_sub_wedges = list_of_remaining_sub_wedges[1:]
            sub_wedge_merged = mergeTwoSubWedgesAdjascentInRotationAxis(
                current_sub_wedge, next_sub_wedge
            )
            if sub_wedge_merged is None:
                list_of_merged_sub_wedges.append(current_sub_wedge)
                current_sub_wedge = next_sub_wedge
            else:
                current_sub_wedge = sub_wedge_merged
        list_of_merged_sub_wedges.append(current_sub_wedge)
    return list_of_merged_sub_wedges
